@@ -153,39 +153,59 @@ bus_to_str (WacomBusType bus)
 }
 
 char *
-make_match_string (const char *name, WacomBusType bus, int vendor_id, int product_id)
+make_match_string (const char *name, const char *uniq, WacomBusType bus, int vendor_id, int product_id)
 {
-	return g_strdup_printf("%s:%04x:%04x%s%s",
+	return g_strdup_printf("%s|%04x|%04x%s%s%s%s",
 				bus_to_str (bus),
 				vendor_id, product_id,
-				name ? ":" : "",
-				name ? name : "");
+				(name || uniq) ? "|" : "",
+				name ? name : "",
+			        uniq ? "|" : "",
+			        uniq ? uniq : "");
 }
 
 static gboolean
-match_from_string(const char *str, WacomBusType *bus, int *vendor_id, int *product_id, char **name)
+match_from_string(const char *str_in, WacomBusType *bus, int *vendor_id, int *product_id, char **name, char **uniq)
 {
-	int rc = 1, len = 0;
-	char busstr[64];
+	gboolean rc = FALSE;
+	guint64 num;
+	char *str = g_strdup(str_in);
+	char **components = NULL;
 
-	rc = sscanf(str, "%63[^:]:%x:%x:%n", busstr, vendor_id, product_id, &len);
-	if (len > 0) {
-		/* Grumble grumble scanf handling of %n */
-		*name = g_strdup(str+len);
-	} else if (rc == 3) {
-		*name = NULL;
-	} else {
-		return FALSE;
+	if (g_str_has_suffix(str_in, ";"))
+		str[strlen(str) - 1] = '\0';
+
+	components = g_strsplit(str, "|", 16);
+	if (!components[0] || !components[1] || !components[2])
+		goto out;
+
+	*bus = bus_from_str (components[0]);
+	if (!g_ascii_string_to_unsigned(components[1], 16, 0, 0xffff, &num, NULL))
+		goto out;
+	*vendor_id = (int)num;
+
+	if (!g_ascii_string_to_unsigned(components[2], 16, 0, 0xffff, &num, NULL))
+		goto out;
+	*product_id = (int)num;
+
+	if (components[3]) {
+		*name = g_strdup(components[3]);
+		if (components[4])
+			*uniq = g_strdup(components[4]);
 	}
-	*bus = bus_from_str (busstr);
 
-	return TRUE;
+	rc = TRUE;
+out:
+	free(str);
+	g_strfreev(components);
+	return rc;
 }
 
 static WacomMatch *
 libwacom_match_from_string(const char *matchstr)
 {
 	char *name = NULL;
+	char *uniq = NULL;
 	int vendor_id, product_id;
 	WacomBusType bus;
 	WacomMatch *match;
@@ -195,16 +215,18 @@ libwacom_match_from_string(const char *matchstr)
 
 	if (g_str_equal(matchstr, GENERIC_DEVICE_MATCH)) {
 		name = NULL;
+		uniq = NULL;
 		bus = WBUSTYPE_UNKNOWN;
 		vendor_id = 0;
 		product_id = 0;
-	} else if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name)) {
+	} else if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name, &uniq)) {
 		DBG("failed to match '%s' for product/vendor IDs. Skipping.\n", matchstr);
 		return NULL;
 	}
 
-	match = libwacom_match_new(name, bus, vendor_id, product_id);
+	match = libwacom_match_new(name, uniq, bus, vendor_id, product_id);
 	free(name);
+	free(uniq);
 
 	return match;
 }
@@ -213,19 +235,21 @@ static gboolean
 libwacom_matchstr_to_paired(WacomDevice *device, const char *matchstr)
 {
 	char *name = NULL;
+	char *uniq = NULL;
 	int vendor_id, product_id;
 	WacomBusType bus;
 
 	g_return_val_if_fail(device->paired == NULL, FALSE);
 
-	if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name)) {
+	if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name, &uniq)) {
 		DBG("failed to match '%s' for product/vendor IDs. Ignoring.\n", matchstr);
 		return FALSE;
 	}
 
-	device->paired = libwacom_match_new(name, bus, vendor_id, product_id);
+	device->paired = libwacom_match_new(name, uniq, bus, vendor_id, product_id);
 
 	free(name);
+	free(uniq);
 	return TRUE;
 }
 
@@ -384,7 +408,9 @@ static const struct {
 	{ "Ring2", WACOM_BUTTON_RING2_MODESWITCH },
 	{ "Touchstrip", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH },
 	{ "Touchstrip2", WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH },
-	{ "OLEDs", WACOM_BUTTON_OLED }
+	{ "OLEDs", WACOM_BUTTON_OLED },
+	{ "Dial", WACOM_BUTTON_DIAL_MODESWITCH },
+	{ "Dial2", WACOM_BUTTON_DIAL2_MODESWITCH },
 };
 
 static const struct {
@@ -647,6 +673,8 @@ libwacom_parse_buttons(WacomDevice *device,
 	device->ring_num_modes = libwacom_parse_num_modes(device, keyfile, "RingNumModes", WACOM_BUTTON_RING_MODESWITCH);
 	device->ring2_num_modes = libwacom_parse_num_modes(device, keyfile, "Ring2NumModes", WACOM_BUTTON_RING2_MODESWITCH);
 	device->strips_num_modes = libwacom_parse_num_modes(device, keyfile, "StripsNumModes", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH);
+	device->dial_num_modes = libwacom_parse_num_modes(device, keyfile, "DialNumModes", WACOM_BUTTON_DIAL_MODESWITCH);
+	device->dial2_num_modes = libwacom_parse_num_modes(device, keyfile, "Dial2NumModes", WACOM_BUTTON_DIAL2_MODESWITCH);
 }
 
 static void
@@ -730,12 +758,6 @@ libwacom_parse_features(WacomDevice *device, GKeyFile *keyfile)
 	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Touch", NULL))
 		device->features |= FEATURE_TOUCH;
 
-	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Ring", NULL))
-		device->features |= FEATURE_RING;
-
-	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Ring2", NULL))
-		device->features |= FEATURE_RING2;
-
 	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Reversible", NULL))
 		device->features |= FEATURE_REVERSIBLE;
 
@@ -747,15 +769,13 @@ libwacom_parse_features(WacomDevice *device, GKeyFile *keyfile)
 	    device->features & FEATURE_REVERSIBLE)
 		g_warning ("Tablet '%s' is both reversible and integrated in screen. This is impossible", libwacom_get_match(device));
 
-	if (!(device->features & FEATURE_RING) &&
-	    (device->features & FEATURE_RING2))
-		g_warning ("Tablet '%s' has Ring2 but no Ring. This is impossible", libwacom_get_match(device));
-
 	if (!(device->features & FEATURE_TOUCH) &&
 	    (device->features & FEATURE_TOUCHSWITCH))
 		g_warning ("Tablet '%s' has touch switch but no touch tool. This is impossible", libwacom_get_match(device));
 
+	device->num_rings = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumRings", NULL);
 	device->num_strips = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumStrips", NULL);
+	device->num_dials = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumDials", NULL);
 
 	string_list = g_key_file_get_string_list(keyfile, FEATURES_GROUP, "StatusLEDs", NULL, NULL);
 	if (string_list) {
@@ -869,7 +889,7 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 	}
 
 	layout = g_key_file_get_string(keyfile, DEVICE_GROUP, "Layout", NULL);
-	if (layout) {
+	if (layout && layout[0] != '\0') {
 		/* For the layout, we store the full path to the SVG layout */
 		device->layout = g_build_filename (datadir, "layouts", layout, NULL);
 		g_free (layout);
@@ -892,6 +912,7 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 	}
 
 	device->num_strips = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumStrips", NULL);
+	device->num_dials = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumDials", NULL);
 	device->buttons = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 						NULL, g_free);
 	device->status_leds = g_array_new (FALSE, FALSE, sizeof(WacomStatusLEDs));
